-----------------------------------------------------------------------------------------------------------------------------------------
-- Entity Name:  GrayCnt4b
-- File Name:    GrayCnt4b.vhd
-- File Path:    E:\Projects\GrayCounter\Vhdl
-- Project :	Fun
--
-- Purpose: 	Provide a 4 bit Gray counter with two databit
--				outputs. One from the incrementer LUT and one from the 
--				counter register. This is easy for BRAM addressing purposes
-- 				when used as Self Addressing Fifo.
--				A Terminal Count is also available.
--
--				The counter is RLOC-ed in the HDL.
--				This works excellent in Mentor Precision Tools.
--
-- Limitations: Synthesis tool Mentor Precision.
--				Don't know about Synplicity.
--
-- Authors: Marc Defossez
--
-- Tools: 
--
--  Contact:    e-mail  hotline@xilinx.com        phone   + 1 800 255 7778
--
--  Disclaimer:
--              LIMITED WARRANTY AND DISCLAMER.
--              These designs are provided to you "as is". Xilinx and its licensors make and you
--              receive no warranties or conditions, express, implied, statutory or otherwise, and
--              Xilinx specifically disclaims any implied warranties of merchantability, non-infringement,
--              or fitness for a particular purpose. Xilinx does not warrant that the functions contained
--              in these designs will meet your requirements, or that the operation of these designs will
--              be uninterrupted or error free, or that defects in the designs will be corrected.
--              Furthermore, Xilinx does not warrant or make any representations regarding use or the
--              results of the use of the designs in terms of correctness, accuracy, reliability, or otherwise.
--
--              LIMITATION OF LIABILITY.
--              In no event will Xilinx or its licensors be liable for any loss of data, lost profits, cost or
--              procurement of substitute goods or services, or for any special, incidental, consequential,
--              or indirect damages arising from the use or operation of the designs or accompanying
--              documentation, however caused and on any theory of liability. This limitation will apply
--              even if Xilinx has been advised of the possibility of such damage. This limitation shall
--              apply not-withstanding the failure of the essential purpose of any limited remedies herein.
--
--              Copyright  2002 Xilinx, Inc.
--              All rights reserved
--
-- Revision History:       Created:       20/02/04
--                         Last opened:   Friday,Feb 20 04
--
----------------------------------------------------------------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_UNSIGNED.all;
-- Simulation libraries
	-- synopsys translate_off
	library UNISIM;
	use UNISIM.all;
	-- synopsys translate_on
-- end simulation libraries

-- Libraries for Synplicity
-- This library setting is needed for the instantiated LUT.
-- When this library is not mentionned Synplify_Pro 7.3.4 fails during synthesis of
-- the hardware with an error about the INIT value.
-- library virtex2;
-- use virtex2.components.all;
-- This library is ignored by Mentor Precision but must be commented
-- for simulation with ModelSim.
-- Only a warning is generated.
-- End libraries for Synplicity

entity GrayCnt4b is
    port (
    	Clk		: in std_logic;
        Rst		: in std_logic;
        Ena		: in std_logic;
        CmpTc4b		: out std_logic;
        FfTc4b		: out std_logic;
        IncOut4b	: out std_logic_vector (3 downto 0);
        CntOut4b	: out std_logic_vector (3 downto 0)
    );
end  GrayCnt4b ;

architecture GrayCnt4b_arch of GrayCnt4b  is
----------------------------------------------------------------------------------------------------------------------------------------
-- Component Instantiation
----------------------------------------------------------------------------------------------------------------------------------------
component LUT4
generic (
	INIT : bit_vector (15 downto 0)
);
port (
	I0	: in std_Logic;
	I1	: in std_Logic;
	I2	: in std_Logic;
	I3	: in std_Logic;
	O	: out std_Logic
);
end component;

component FDCE
port (
	D		: in std_logic;
	CE		: in std_logic;
	C		: in std_logic;
	CLR		: in std_logic;
	Q		: out std_logic
);
end component;
----------------------------------------------------------------------------------------------------------------------------------------
-- Signal Declarations
----------------------------------------------------------------------------------------------------------------------------------------
type InitIncVal is array (3 downto 0) of bit_vector (15 downto 0);
constant IncLutVal : InitIncVal := (
	X"FE10",	-- Address three
	X"B0F4",
	X"E44E",
	X"3CC3"		-- Address zero
);
constant TcLutVal : bit_vector (15 downto 0) := X"0001";
--
-- The INITialisation values for the LUTs:
--  1111111000010000	X"FE10"
--	1011000011110100	X"B0F4"
--	1110010001001110	X"E44E"
--	0011110011000011	X"3CC3"
--
signal High 			: std_logic;
signal Low	 			: std_logic;
signal IntDatOut		: std_logic_vector (3 downto 0);
signal IntIncOut 		: std_logic_vector (3 downto 0);
signal IntCmpTc			: std_logic;
signal IntFfTc			: std_logic;

-- Synplicity attributes
attribute SYN_HIER : string;
	attribute SYN_HIER of GrayCnt4b_arch : architecture is "hard";
-- Xilinx attributes
-- Needed to set a range from 4 downto 1 (in stead of from 3 downto 0) because
-- else Synplicity give a serious compiler error.
-- For Precision a string does not need to get a range.
attribute RLOC : string;
	type RlocValue is array (3 downto 0) of string (4 downto 1);
	constant RlocVal : RlocValue := (
		"X0Y1",		-- Bit 3 (LUT and FF)
		"X0Y1",		-- Bit 2
		"X0Y0",		-- Bit 1
		"X0Y0"		-- Bit 0
	);
	attribute RLOC of TcLut : label is "X0Y2";
	attribute RLOC of TcFf : label is "X0Y2";
----------------------------------------------------------------------------------------------------------------------------------------
--
begin

High <= '1';
Low <= '0';

GenLut : for i in 0 to 3 generate
	attribute RLOC of IncrLut : label is RlocVal(i);
	attribute RLOC of IncrFf : label is RlocVal(i);
	begin
	IncrLut : LUT4
	generic map (
		INIT => IncLutVal(i)
	)
	port map (
		I0 	=> IntDatOut(0),
		I1 	=> IntDatOut(1),
		I2 	=> IntDatOut(2),
		I3 	=> IntDatOut(3),
		O 	=> IntIncOut(i)
	);
	IncrFf : FDCE
	port map (
		D	=> IntIncOut(i),
		CE	=> Ena,
		C	=> Clk,
		CLR	=> Rst,
		Q	=> IntDatOut(i)
	);
end generate;
CntOut4b <= IntDatOut;
IncOut4b <= IntIncOut;

TcLut : LUT4
	generic map (
		INIT => TcLutVal
	)
	port map (
		I0 	=> IntIncOut(0),
		I1 	=> IntIncOut(1),
		I2 	=> IntIncOut(2),
		I3 	=> IntIncOut(3),
		O 	=> IntCmpTc
	);
CmpTc4b <= IntCmpTc;
TcFf : FDCE
	port map (
		D	=> IntCmpTc,
		CE	=> Ena,
		C	=> Clk,
		CLR	=> Rst,
		Q	=> IntFfTc
	);
FfTc4b <= IntFfTc;

end  GrayCnt4b_arch ;