-----------------------------------------------------------------------------------------------------------------------------------------
-- Entity Name:  AltEmifInt
-- File Name:    AltEmifInt.vhd
-- File Path:    E:\Projects\EmifTi\Design\Vhdl\
-- Project :	Connect a TI DSP via the EMIF interface to a Xilinx FPGA.
--
-- Purpose: 	Provide a fast memory based FPGA interface to the EMIF bus of a TI C64x DSP.
--				This interface is using Block Select RAM FPGA memory.
--				The RAM's PORT A is connected for read and write to the EMIF, while the port B
--				can be used by any back-end design in the FPGA.
--				Different Block SelectRAM device combinations can be used. the interface speed
--				will depend the BlockRAM solution chosen.
--				This application is using one Block SelectRAM in 32-bit mode (256 deep).
--						Chosen for this solution because it nicely can be hooked to an
--						internal running PPC or Microblaze processor.
--
--				The memeory size can be made larger when other memeory combinations are chosen.
--				for a 32-bit interface, two Block SelectRAM in 16-bit mode can be chosen.
--				They will occupy the same design area in the FPGA and the interface will
--				run at the same speed.
--				For a 64-bit interface two 30-bit or four 16-bit memories can be used.
--				This design will run somewhat slower.
--
--				BlockRAM parameters and EMIF parameters are entered as VHDL 'generic' statements.
--
--				The design is implemented with ISE 6.2sp1 in a XC2VP20-FF896-6 part.
--
-- Limitations: none
--
-- Authors: Marc Defossez
--
-- Tools: 
--
--  Contact:    e-mail  hotline@xilinx.com        phone   + 1 800 255 7778
--
--  Disclaimer:
--              LIMITED WARRANTY AND DISCLAMER.
--              These designs are provided to you "as is". Xilinx and its licensors make and you
--              receive no warranties or conditions, express, implied, statutory or otherwise, and
--              Xilinx specifically disclaims any implied warranties of merchantability, non-infringement,
--              or fitness for a particular purpose. Xilinx does not warrant that the functions contained
--              in these designs will meet your requirements, or that the operation of these designs will
--              be uninterrupted or error free, or that defects in the designs will be corrected.
--              Furthermore, Xilinx does not warrant or make any representations regarding use or the
--              results of the use of the designs in terms of correctness, accuracy, reliability, or otherwise.
--
--              LIMITATION OF LIABILITY.
--              In no event will Xilinx or its licensors be liable for any loss of data, lost profits, cost or
--              procurement of substitute goods or services, or for any special, incidental, consequential,
--              or indirect damages arising from the use or operation of the designs or accompanying
--              documentation, however caused and on any theory of liability. This limitation will apply
--              even if Xilinx has been advised of the possibility of such damage. This limitation shall
--              apply not-withstanding the failure of the essential purpose of any limited remedies herein.
--
--              Copyright  2002 Xilinx, Inc.
--              All rights reserved
--
-- Revision History:       Created:       15/03/04
--                         Last opened:   Monday,Mar 15 04
--
----------------------------------------------------------------------------------------------------------------------------------------
--
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_UNSIGNED.all;
-- Simulation libraries
	-- synopsys translate_off
	library UNISIM;
	use UNISIM.all;
	-- synopsys translate_on
-- end simulation libraries

-- Libraries for Synplicity
-- Neede for componnets as LUT, SRL which get an INIT parameter.
-- When this library is not mentionned Synplify_Pro fails during synthesis of
-- the hardware with an error about the INIT value.
-- library virtex2;
-- use virtex2.components.all;
-- This library is ignored by Mentor Precision.
-- Only a warning is generated.
-- End libraries for Synplicity

entity AltEmifInt is
	generic (
		-- EMIF port / Port A
		NumberOfDataBits 	: integer := 32;	-- 32-bit EMIF
		NumberOfAddrBits 	: integer := 9; 	-- 9 Address lines for Block RAM
		-- System Port / Port B
		NumberOfSysDataBits : integer := 32;
		NumberOfSysAddrBits : integer := 9;
		-- Block SelectRAM
		UsedBlockRamWidth 	: integer := 32
	);
    port (
        -- EMIF side
        EmifClk		: in std_logic;
        EmifRst		: in std_logic;
        EmifAddr	: in std_logic_vector (NumberOfAddrBits-1 downto 0);
        EmifData	: inout std_logic_vector (NumberOfDataBits-1 downto 0);
        EmifSoe		: in std_logic;
        EmifCe		: in std_logic;
        EmifSwe		: in std_logic;
        EmifSads	: in std_logic;
        EmifBe		: in std_logic_vector (((NumberOfDataBits/8)-1) downto 0);
        -- Internal or system side
        SysClk		: in std_logic;
        SysRst		: in std_logic;
        SysAddr		: in std_logic_vector (NumberOfSysAddrBits-1 downto 0);     -- Address can be different width.
        SysData		: Inout std_logic_vector (NumberOfSysDataBits-1 downto 0);  -- Data can be different width.
        SysRdWr		: in std_logic;
        SysEna		: in std_logic_vector (((NumberOfSysDataBits/UsedBlockRamWidth)-1) downto 0)
    );
end  AltEmifInt ;

architecture AltEmifInt_arch of AltEmifInt  is
----------------------------------------------------------------------------------------------------------------------------------------
-- Component Instantiation
----------------------------------------------------------------------------------------------------------------------------------------
-- Single ended clock input buffer primitive
component IBUFG_LVCMOS33
	port (
		I : in std_logic;
		O : out std_logic
	);
end component;

component DCM
--pragma translate_off
	generic (
		CLKIN_DIVIDE_BY_2 	: boolean ;
 		DESKEW_ADJUST 		: string ;
		CLKDV_DIVIDE 		: real ;
		CLKOUT_PHASE_SHIFT 	: string ;
		PHASE_SHIFT 		: integer
		);
--pragma translate_on
	port (
		CLKIN   	: in  std_logic ;
		CLKFB   	: in  std_logic ;
		DSSEN 		: in  std_logic ;
		PSINCDEC	: in  std_logic ;
		PSEN 		: in  std_logic ;
		PSCLK 		: in  std_logic ;
		RST     	: in  std_logic ;
		CLK0    	: out std_logic ;
		CLK90   	: out std_logic ;
		clk180  	: out std_logic ;
		CLK270  	: out std_logic ;
		CLK2X   	: out std_logic ;
		CLK2X180	: out std_logic ;
		CLKDV   	: out std_logic ;
		CLKFX   	: out std_logic ;
		CLKFX180	: out std_logic ;
		LOCKED  	: out std_logic ;
		PSDONE  	: out std_logic ;
		STATUS  	: out std_logic_vector(7 downto 0)); 
end component;

component BUFG
	port (
		I : in std_logic;
		O : out std_logic
	);
end component;
--
-- Only one of the two instantiated components is used.
-- The design as it is now is using ONE block selectRAM 512x32.
-- When the other BRAM is used, please addapt the generic values
-- at the top of the design.
--
component RAMB16_S18_S18 -- Block RAM of 16+2 * 1024
	port (
		WEA		: in std_logic;
		ENA		: in std_logic;
		SSRA	: in std_logic;
		CLKA	: in std_logic;
		ADDRA	: in std_logic_vector (9 downto 0);
		DIA		: in std_logic_vector (15 downto 0);
		DIPA	: in std_logic_vector (1 downto 0);
		DOPA	: out std_logic_vector (1 downto 0);
		DOA		: out std_logic_vector (15 downto 0);
		
		WEB		: in std_logic;
		ENB		: in std_logic;
		SSRB	: in std_logic;
		CLKB	: in std_logic;
		ADDRB	: in std_logic_vector (9 downto 0); 
		DIB		: in std_logic_vector (15 downto 0);
		DIPB	: in std_logic_vector (1 downto 0); 
		DOPB	: out std_logic_vector (1 downto 0); 
		DOB		: out std_logic_vector (15 downto 0)
	);
end component;

component RAMB16_S36_S36 -- Block RAM of 32+4 * 512
	port (
		WEA		: in std_logic;
		ENA		: in std_logic;
		SSRA	: in std_logic;
		CLKA	: in std_logic;
		ADDRA	: in std_logic_vector (8 downto 0);
		DIA		: in std_logic_vector (31 downto 0);
		DIPA	: in std_logic_vector (3 downto 0);
		DOPA	: out std_logic_vector (3 downto 0);
		DOA		: out std_logic_vector (31 downto 0);
		
		WEB		: in std_logic;
		ENB		: in std_logic;
		SSRB	: in std_logic;
		CLKB	: in std_logic;
		ADDRB	: in std_logic_vector (8 downto 0); 
		DIB		: in std_logic_vector (31 downto 0);
		DIPB	: in std_logic_vector (3 downto 0); 
		DOPB	: out std_logic_vector (3 downto 0); 
		DOB		: out std_logic_vector (31 downto 0)
	);
end component;

component IBUF_LVCMOS33
	port (
		I : in std_logic;
		O : out std_logic
	);
end component;
component IOBUF_LVCMOS33_F_24
	port (
		T : in std_logic;
		I : in std_logic;
		O : out std_logic;
		IO : inout std_logic
	);
end component;
----------------------------------------------------------------------------------------------------------------------------------------
-- Signal Declarations
----------------------------------------------------------------------------------------------------------------------------------------
signal Low		: std_logic;
signal MLow	: std_logic_vector (((UsedBlockRamWidth/8)-1) downto 0);
signal High		: std_logic;
-- EMIF side
signal IntEmifClkToDcm	: std_logic;
signal IntEmifClk		: std_logic;
signal IntEmifRst		: std_logic;
signal IntDcmClkToBufg  : std_logic;
signal IntLocked        : std_logic;

signal IntEmifSoe			: std_logic;
signal IntEmifDataIn		: std_logic_vector (NumberOfDataBits-1 downto 0);
signal IntEmifDataOut		: std_logic_vector (NumberOfDataBits-1 downto 0);
signal IntEmifSads			: std_logic;
signal IntEmifSwe			: std_logic;
signal IntEmifCe			: std_logic;
signal IntEmifBe			: std_logic_vector (((NumberOfDataBits/8)-1) Downto 0);
signal IntWritePortA		: std_logic;
signal IntEnaPortA       	: std_logic_vector (((NumberOfDataBits/UsedBlockRamWidth)-1) downto 0);  --(((64/16)-1) downto 0)
signal IntRstPortA      	: std_logic;
signal IntClkPortA          : std_logic;
signal IntAddrPortA         : std_logic_vector (NumberOfAddrBits-1 downto 0);
signal IntDataInPortA		: std_logic_vector (NumberOfDataBits-1 downto 0);
signal IntDataOutPortA		: std_logic_vector (NumberOfDataBits-1 downto 0);
-- System side
signal IntWritePortB        : std_logic;
signal IntEnaPortB			: std_logic_vector (((NumberOfSysDataBits/UsedBlockRamWidth)-1) downto 0);
signal IntRstPortB			: std_logic;
signal IntClkPortB          : std_logic;
signal IntAddrPortB         : std_logic_vector (NumberOfSysAddrBits-1 downto 0);
signal IntDataInPortB       : std_logic_vector (NumberOfSysDataBits-1 downto 0);
signal IntDataOutPortB      : std_logic_vector (NumberOfSysDataBits-1 downto 0);
signal IntSysEna			: std_logic_vector ((NumberOfSysDataBits/UsedBlockRamWidth)-1 downto 0);
signal IntSysRdWr			: std_logic;
signal IntSysDataIn			: std_logic_vector (NumberOfSysDataBits-1 downto 0);
signal IntSysDataOut		: std_logic_vector (NumberOfSysDataBits-1 downto 0);
signal IntSysAddr			: std_logic_vector (NumberOfSysAddrBits-1 downto 0);
signal IntIntSysClk			: std_logic;
signal IntSysClk			: std_logic;
signal IntSysRst			: std_logic;
--
-- Xilinx constraints.
--
attribute DLL_FREQUENCY_MODE 	: string;
attribute CLKOUT_PHASE_SHIFT 	: string;
attribute PHASE_SHIFT 			: string;
attribute CLKDV_DIVIDE 			: string;
attribute DUTY_CYCLE_CORRECTION : string;
attribute DESKEW_ADJUST		 	: string;
attribute CLKIN_DIVIDE_BY_2 	: string;

attribute DLL_FREQUENCY_MODE 	of EmifDcm	: label is "LOW" ;
attribute CLKOUT_PHASE_SHIFT 	of EmifDcm	: label is "NONE" ;
attribute PHASE_SHIFT        	of EmifDcm	: label is "0" ;
attribute DUTY_CYCLE_CORRECTION of EmifDcm	: label is "TRUE" ;
attribute CLKDV_DIVIDE          of EmifDcm	: label is "4" ;
attribute DESKEW_ADJUST         of EmifDcm	: label is "2" ;
attribute CLKIN_DIVIDE_BY_2 	of EmifDcm	: label is "FALSE" ;
-- Do not place the registers at, at the back side (Port B), of the Block SelectRAM in
-- the IOB. This to show the back end side of the design in the FPGA.
attribute IOB : string;
	attribute IOB of IntAddrPortB : signal is "FALSE";
	attribute IOB of IntDataInPortB : signal is "FALSE";
	attribute IOB of IntSysDataOut : signal is "FALSE";
----------------------------------------------------------------------------------------------------------------------------------------
--
begin
--
Low 	<= '0' ;
MultiLow : for l in ((UsedBlockRamWidth/8)-1) downto 0 generate
	Mlow(l) <= Low;
end generate;
High 	<= '1' ;
------------------------------------------------------------------------
-- FPGA EMIF Clock generation.
------------------------------------------------------------------------
EmifIbufg : IBUFG_LVCMOS33
	port map (
		I 	=> EmifClk,
		O 	=> IntEmifClkToDcm
	);
EmifDcm : DCM	
--pragma translate_off
	generic map (
	CLKIN_DIVIDE_BY_2 	=> FALSE,
 	DESKEW_ADJUST 		=> "0",
	CLKDV_DIVIDE		=> 4.0,
	CLKOUT_PHASE_SHIFT 	=> "NONE",
	PHASE_SHIFT 		=> 0
		) 
--pragma translate_on
	port map (
		CLKIN   	=> IntEmifClkToDcm,
		CLKFB   	=> IntEmifClk,
		DSSEN 		=> Low,
		PSINCDEC	=> Low,
		PSEN 		=> Low,
		PSCLK 		=> Low,
		RST     	=> IntEmifRst,
		CLK0    	=> IntDcmClkToBufg,
		CLK90   	=> open,
		clk180  	=> open,
		CLK270  	=> open,
		CLK2X   	=> open,
		CLK2X180	=> open,
		CLKDV   	=> open,
		CLKFX   	=> open,
		CLKFX180	=> open,
		LOCKED  	=> IntLocked,
		PSDONE  	=> open,
		STATUS  	=> open
		);
EmifBufg : BUFG
	port map (
		I => IntDcmClkToBufg,
		O => IntEmifClk
	);
------------------------------------------------------------------------
-- Emif side logic.
------------------------------------------------------------------------
-- Address Inputs
EmifAddrGen : for a in NumberOfAddrBits-1 downto 0 generate
	InBuf : IBUF_LVCMOS33
	port map (
		I => EmifAddr(a),
		O => IntAddrPortA(a)
	);
end generate;
-- Control Inputs
EmifSoeBuf : IBUF_LVCMOS33
	port map (
		I => EmifSoe,		-- High = Emif write to Bram, Low = Emif read from Bram
		O => IntEmifSoe
	);
--
EmifRstBuf : IBUF_LVCMOS33
	port map (
		I => EmifRst,		-- High is reset active.
		O => IntEmifRst
	);
--
EmifRdBuf : IBUF_LVCMOS33
	port map (
		I => EmifSwe,		-- Read / write bit. High = read, Low = write
		O => IntEmifSwe
	);
--
EmifSadsBuf : IBUF_LVCMOS33
	port map (
		I => EmifSads,		-- Low is address and data active.
		O => IntEmifSads
	);
--
EmifCeBuf : IBUF_LVCMOS33
	port map (
		I => EmifCe,		-- active low, for use invert the signal.
		O => IntEmifCe
	);
--
EmifBeGen	: for e in ((NumberOfDataBits/8)-1) downto 0 generate
	Ibuf : IBUF_LVCMOS33
	port map (
		I => EmifBe(e),
		O => IntEmifBe(e)
	);
end generate;

-- Emif data width = 64 and Bram data width = 32
-- Emif BE bits = 64/8 = 8
-- Bram's used = 2, thus 2 Bram enables.
-- 8 BE bits must match 2 Bram enables.
-- Needed two gates (AND) of 4 input bits. AND gates because all bits must be 1 before a write
-- or read can happen.
--  64 Emif bits / 32 Bram bits = 2 --> gates
--  64 Emif bits / 8 = 8 BE bits --> divided by two gates = 4 bits per gate.
--
-- Example for 32 Emif bits and 32 Bram bits.
-- 32 Emif / 8 = 4 BE bits
-- 32 Emif bits / 32 Bram bits = 1 --> Gates
-- 4 BE bits / 1 gates = 4 inputs gate.
--
-- Exanple for 63 Emif bits and 16 Bram bits
-- 64 Emif bits / 16 Bram bits = 4 --> gates
-- 64 Emif Bits / 8 = 8 BE bits.
-- 8 BE bits / 4 gates = 2 inputs per gate.
--
EnGateGen : for b in ((NumberOfDataBits/UsedBlockRamWidth)-1) downto 0 generate  -- loop for number of gates
	GateIn4 : if (NumberOfDataBits/8)/(NumberOfDataBits/UsedBlockRamWidth) = 4 generate -- loop for number of gate inputs
		IntEnaPortA(b) <= not IntEmifCe and (IntEmifBe((4*b)+3) and IntEmifBe((4*b)+2) and IntEmifBe((4*b)+1) and IntEmifBe(4*b));
	end generate;
	GateIn2 : if (NumberOfDataBits/8)/(NumberOfDataBits/UsedBlockRamWidth) = 2 generate
		IntEnaPortA(b) <= not IntEmifCe and (IntEmifBe((2*b)+1) and IntEmifBe(2*b));
	end generate;
	GateIn1 : if (NumberOfDataBits/8)/(NumberOfDataBits/UsedBlockRamWidth) = 1 generate
		IntEnaPortA(b) <= not IntEmifCe and IntEmifBe(b);
	end generate;
end generate;
--
-- Data Inputs
EmifDataGen : for d in NumberOfDataBits-1 downto 0 generate
	IoBuf : IOBUF_LVCMOS33_F_24
		port map (
			T => Low,
			I => IntEmifDataOut(d),
			O => IntEmifDataIn(d),
			IO => EmifData(d)
		);
	IntDataInPortA(d) <= IntEmifSoe and IntEmifDataIn(d);
	IntEmifDataOut(d) <= not IntEmifSoe and IntDataOutPortA(d);
end generate;
------------------------------------------------------------------------
-- Block SelectRAM bank
------------------------------------------------------------------------
IntClkPortA		<= IntEmifClk;
IntRstPortA		<= IntEmifRst;
 -- SADS is active Low and SWE is cative low write. BRAM is active high write.
IntWritePortA	<= not IntEmifSads and not IntEmifSwe ;

-- Use the correct Block SelectRAM instantiation!
The : for n in ((NumberOfDataBits/UsedBlockRamWidth)-1) downto 0 generate
--	Ram : RAMB16_S18_S18 -- Block RAM of 16+2*1024
	Ram : RAMB16_S36_S36 -- Block RAM of 32+4*512
	port map (
		WEA		=> IntWritePortA,
		ENA		=> IntEnaPortA(n),
		SSRA	=> IntRstPortA,
		CLKA	=> IntClkPortA,
		ADDRA	=> IntAddrPortA,
		DIA		=> IntDataInPortA (((((UsedBlockRamWidth)-1)*(1+n))+n) downto ((((UsedBlockRamWidth)-1)*n)+n)),
		DIPA	=> Mlow,
		DOPA	=> open,
		DOA		=> IntDataOutPortA (((((UsedBlockRamWidth)-1)*(1+n))+n) downto ((((UsedBlockRamWidth)-1)*n)+n)),

		WEB		=> IntWritePortB,
		ENB		=> IntEnaPortB(n),
		SSRB	=> IntRstPortB,
		CLKB	=> IntClkPortB,
		ADDRB	=> IntAddrPortB,
		DIB		=> IntDataInPortB (((((UsedBlockRamWidth)-1)*(1+n))+n) downto ((((UsedBlockRamWidth)-1)*n)+n)),
		DIPB	=> Mlow,
		DOPB	=> open,
		DOB		=> IntDataOutPortB (((((UsedBlockRamWidth)-1)*(1+n))+n) downto ((((UsedBlockRamWidth)-1)*n)+n))
	);
end generate;

IntClkPortB		<= IntSysClk;
IntRstPortB		<= IntSysRst;
IntWritePortB	<= not IntSysRdWr;
IntEnaPortB		<= IntSysEna;
------------------------------------------------------------------------
-- System Side Logic.
------------------------------------------------------------------------
-- All registers have been defined appart because of floorplanning.
--
-- Address Registers and IO.
SysAddrGen : for a in (NumberOfSysAddrBits-1) downto 0 generate
	InBuf : IBUF_LVCMOS33
	port map (
		I => SysAddr(a),
		O => IntSysAddr(a)
	);
end generate;
--
SysAddrReg : process (IntSysClk, IntSysRst, IntSysAddr)
begin
	if (IntSysRst = '1') then
		IntAddrPortB <= (others => '0');
	elsif (IntSysClk'event and IntSysClk = '1') then
		IntAddrPortB <= IntSysAddr;
	end if;
end process;
-- Data Registers and IO
SysDataGen : for d in (NumberOfSysDataBits-1) downto 0 generate
	IoBuf : IOBUF_LVCMOS33_F_24
	port map (
		T 	=> Low,
		I 	=> IntSysDataOut(d),	-- input of the IOBUF.
		O 	=> IntSysDataIn(d),		-- output of the IOBUF.
		IO	=> SysData(d)  			-- Output to the outside world.
	);
end generate;
--
SysInDataReg : process (IntSysClk, IntSysRst, IntSysDataIn)
begin
	if (IntSysRst = '1') then
		IntDataInPortB <= (others => '0');
	elsif (IntSysClk'event and IntSysClk = '1') then
		if (IntSysRdWr = '0') then
			IntDataInPortB <= IntSysDataIn;
		end if;
	end if;
end process;
--
SysOutDataReg : process (IntSysClk, IntSysRst, IntSysDataOut)
begin
	if (IntSysRst = '1') then
		IntSysDataOut <= (others => '0');
	elsif (IntSysClk'event and IntSysClk = '1') then
		if (IntSysRdWr = '1') then
			IntSysDataOut <= IntDataOutPortB;
		end if;
	end if;
end process;
--
-- Control registers and IO
SysRdWrBuf : IBUF_LVCMOS33
	port map (
		I => SysRdWr,
		O => IntSysRdWr -- '1' = Read
	);
--
SysEnaBuf : for n in ((NumberOfSysDataBits/UsedBlockRamWidth)-1) downto 0 generate
	Buf : IBUF_LVCMOS33
	port map (
		I => SysEna(n),
		O => IntSysEna(n)	-- '1' = Enabled.
	);
end generate;
--
SysRstBuf : IBUF_LVCMOS33
	port map (
		I => SysRst,
		O => IntSysRst
	);
--
SysClkIbuf : IBUFG_LVCMOS33
	port map (
		I => SysClk,
		O => IntIntSysClk
	);
SysClkBuf : BUFG
	port map (
		I => IntIntSysClk,
		O => IntSysClk
	);
	
------------------------------------------------------------------------
end  AltEmifInt_arch ;